/*
 * Copyright (c) 2010  Nokia Corporation
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "fbdev.h"
#include "sgx_pvr2d.h"
#include "sgx_dri2.h"
#include "perf.h"

#include "exa.h"
#include "x-hash.h"

/* PVR2D memory can only be freed once all PVR2D operations using it have
 * completed. In order to avoid waiting for this synchronously, defer freeing
 * of PVR2D memory with outstanding operations until an appropriate time.
 */
static struct PVR2DMemDestroy {
	struct PVR2DPixmap pix;
	struct PVR2DMemDestroy *next;
} *DelayedPVR2DMemDestroy;

static unsigned int page_size;


static void doDestroyMemory(struct PVR2DPixmap *ppix)
{
	CALLTRACE("%s: Start\n", __func__);

#if SGX_CACHE_SEGMENTS
	if (AddToCache (ppix->shmid, ppix->shmaddr, ppix->pvr2dmem,
	                ppix->mallocaddr, ppix->shmsize))
		return;
#endif

	if (ppix->pvr2dmem)
		PVR2DMemFree(pvr2d_get_screen()->context, ppix->pvr2dmem);

	if (ppix->shmid != -1) {
		shmdt(ppix->shmaddr);
		shmctl(ppix->shmid, IPC_RMID, NULL);
		ppix->shmaddr = NULL;
		ppix->shmid = -1;
		PERF_DECREMENT2(shm_bytes, ppix->shmsize);
		PERF_DECREMENT(shm_segments);
	}

	if (ppix->mallocaddr) {
		free(ppix->mallocaddr);
		ppix->mallocaddr = NULL;
		PERF_DECREMENT2(malloc_bytes, ppix->mallocsize);
		PERF_DECREMENT(malloc_segments);
	}
}

void PVR2DDelayedMemDestroy(Bool wait)
{
	while (DelayedPVR2DMemDestroy) {
		struct PVR2DMemDestroy *destroy = DelayedPVR2DMemDestroy;
		Bool complete =
		          QueryBlitsComplete(&destroy->pix, wait) == PVR2D_OK;

		if (complete || wait) {
		/* This should never happen, but in case it does... */
			if (!complete)
				ErrorF ("Freeing PVR2D memory %p despite incomplete blits!"
				        " SGX may lock up...\n", destroy->pix.pvr2dmem);

			doDestroyMemory(&destroy->pix);
			DelayedPVR2DMemDestroy = destroy->next;
			free(destroy);
		} else
			return;
	}
}

void DestroyPVR2DMemory(struct PVR2DPixmap *ppix)
{
	struct PVR2DMemDestroy *destroy;

	if (ppix->pvr2dmem == pvr2d_get_screen()->sys_mem_info)
		return;

	PVR2DDelayedMemDestroy(FALSE);

	/* Can we free PVR2D memory right away? */
	if (!ppix->pvr2dmem || QueryBlitsComplete(ppix, 0) == PVR2D_OK) {
		doDestroyMemory(ppix);
		return;
	}

	/* No, schedule for delayed freeing */
	destroy = malloc(sizeof(*destroy));

	destroy->pix = *ppix;
	destroy->next = DelayedPVR2DMemDestroy;
	DelayedPVR2DMemDestroy = destroy;
}

Bool PVR2DAllocSHM(struct PVR2DPixmap *ppix)
{
	CALLTRACE("%s: Start\n", __func__);

	ppix->shmsize = ALIGN(ppix->shmsize, page_size);

#if SGX_CACHE_SEGMENTS
	if (GetFromCache (&ppix->shmid, &ppix->shmaddr, (void *)&ppix->pvr2dmem,
	                  NULL, ppix->shmsize))
	    if (ppix->shmaddr)
			return TRUE;
#endif

	ppix->shmid = shmget(IPC_PRIVATE, ppix->shmsize, IPC_CREAT | 0666);

	if (ppix->shmid == -1) {
		perror("shmget failed");
		return FALSE;
	}

	/* lock the SHM segment. This prevents swapping out the memory.
	 * Cache flush/invalidate could cause unhandled page fault
	 * if shared memory was swapped out
	 */
	if (0 != shmctl(ppix->shmid, SHM_LOCK, 0))
		ErrorF("shmctl(SHM_LOCK) failed\n");

	ppix->shmaddr = shmat(ppix->shmid, NULL, 0);

	if (!ppix->shmaddr) {
		perror("shmat failed");
		shmctl(ppix->shmid, IPC_RMID, NULL);
		return FALSE;
	}

	PERF_INCREMENT2(shm_bytes, ppix->shmsize);
	PERF_INCREMENT(shm_segments);

	return TRUE;
}

static Bool PVR2DAllocNormal(struct PVR2DPixmap *ppix)
{
#if SGX_CACHE_SEGMENTS
	ppix->shmsize = ALIGN(ppix->shmsize, page_size);

	if (GetFromCache (&ppix->shmid, &ppix->shmaddr, (void *)&ppix->pvr2dmem,
	                  &ppix->mallocaddr, ppix->shmsize))
		return TRUE;
#endif

	ppix->mallocaddr = calloc(1, ppix->mallocsize);
	if (ppix->mallocaddr)
		return TRUE;

	return FALSE;
}

Bool PVR2DAllocatePixmapMem(struct PVR2DPixmap * ppix, int width, int height,
                           int pitch, pointer pPixData)
{
	struct pvr2d_screen *screen = pvr2d_get_screen();

	if (!page_size)
		page_size = getpagesize();

	/* Using PVR2D flipping support is special, there the pixel memory can
	 * come from system surfaces which are not for individual applications
	 * to allocate or dispose. Here one simply does not touch the details,
	 * but leaves the pixmap for later context specific (such as DRI2)
	 * handling.
	 */
	if (ppix->usage_hint == SGX_EXA_CREATE_PIXMAP_FLIP)
		return TRUE;

	DestroyPVR2DMemory(ppix);
	ppix->pvr2dmem = NULL;
	ppix->mallocaddr = NULL;
	ppix->shmaddr = NULL;
	ppix->shmid = -1;

	if (!pPixData) {
		ppix->mallocsize = pitch * height;
		ppix->shmsize = ALIGN(ppix->mallocsize, page_size);

		if (ppix->usage_hint == CREATE_PIXMAP_USAGE_BACKING_PIXMAP) {
			PVR2DAllocSHM(ppix);
		} else {
			PVR2DAllocNormal(ppix);
		}

		if (ppix->shmaddr)
			return TRUE;

		if (ppix->mallocaddr)
			return TRUE;

		if (height == 0 || pitch == 0)
			return TRUE;

		if (PVR2DMemAlloc (screen->context, height * pitch, 4, 0,
		                   &ppix->pvr2dmem) != PVR2D_OK) {
			ppix->pvr2dmem = NULL;
			return FALSE;
		}
	} else if (pPixData == (void *)~0UL) {
		ppix->pvr2dmem = screen->sys_mem_info;
		ppix->shmid = -1;
		ppix->shmaddr = screen->sys_mem_info->pBase;
	}

	return TRUE;
}

Bool pvr2d_dri2_migrate_pixmap(struct PVR2DPixmap *pix)
{
	struct pvr2d_screen *screen = pvr2d_get_screen();

	if (!page_size)
		page_size = getpagesize();

	if (pix->mallocaddr) {
		pix->shmsize = ALIGN(pix->mallocsize, page_size);

		if (!PVR2DAllocSHM(pix))
			return FALSE;

		DebugF("%s: memcpy %d bytes from malloc (%p) to SHM (%p)\n",
		       __func__, pix->mallocsize, pix->mallocaddr, pix->shmaddr);

		memcpy(pix->shmaddr, pix->mallocaddr, pix->mallocsize);
		free(pix->mallocaddr);
		pix->mallocaddr = NULL;
		pix->mallocsize = 0;
	}

	if (!pix->shmaddr && pix->pvr2dmem != screen->sys_mem_info) {
		if (QueryBlitsComplete(pix, 1) != PVR2D_OK || !PVR2DAllocSHM(pix))
			return FALSE;

		DebugF("%s: memcpy %d bytes from PVR (%p) to SHM (%p)\n",
		        __func__, pix->shmsize, pix->pvr2dmem->pBase, pix->shmaddr);

		memcpy(pix->shmaddr, pix->pvr2dmem->pBase, pix->shmsize);
		PVR2DMemFree(screen->context, pix->pvr2dmem);
		pix->pvr2dmem = NULL;
	}

	if (!PVR2DValidate(pix, TRUE)) {
		ErrorF("%s: !PVR2DValidate()\n", __func__);
		return FALSE;
	}

	PVR2DPixmapOwnership_GPU(pix);

	return TRUE;
}
